@@ -9,14 +9,14 @@ use std::ffi::OsStr;
 use std::path::Path;
 use typed_arena::Arena;
 
-use crate::dijkstra::{mark_syntax, toplevel_list};
+use crate::dijkstra::mark_syntax;
 use crate::lines::{
     apply_groups, enforce_length, format_line_num, join_overlapping, lhs_printable_width,
     rhs_printable_width, visible_groups, MaxLine,
 };
 use crate::parse::{find_lang, parse, parse_lines, read_or_die, ConfigDir};
 use crate::style::apply_colors;
-use crate::syntax::{change_positions, set_next};
+use crate::syntax::{change_positions, matching_lines, set_next};
 
 fn term_width() -> Option<usize> {
     term_size::dimensions().map(|(w, _)| w)
@@ -84,15 +84,27 @@ fn main() {
         None => (parse_lines(&arena, &lhs_src), parse_lines(&arena, &rhs_src)),
     };
 
-    let (lhs, rhs) = toplevel_list(&arena, lhs, rhs);
-    set_next(lhs);
-    set_next(rhs);
-    mark_syntax(lhs, rhs);
+    set_next(&lhs, None);
+    set_next(&rhs, None);
+    mark_syntax(lhs.get(0).map(|n| *n), rhs.get(0).map(|n| *n));
 
-    let lhs_positions = change_positions(&lhs_src, &rhs_src, &[lhs]);
-    let rhs_positions = change_positions(&rhs_src, &lhs_src, &[rhs]);
+    let lhs_positions = change_positions(&lhs_src, &rhs_src, &lhs);
+    let rhs_positions = change_positions(&rhs_src, &lhs_src, &rhs);
+
+    let lhs_matched_lines = matching_lines(&lhs);
+
+    let lang_name = match &lang {
+        Some(lang) => lang.name.clone(),
+        None => "plain text".to_string(),
+    };
+    println!("{}", style::header(rhs_path, &lang_name));
 
     let mut groups = visible_groups(&lhs_positions, &rhs_positions);
+    if groups.is_empty() {
+        println!("No changes found.");
+        return;
+    }
+
     for group in &mut groups {
         group.pad(3, lhs_src.max_line(), rhs_src.max_line());
     }
@@ -117,17 +129,13 @@ fn main() {
     let lhs_colored = apply_colors(&lhs_src, true, &lhs_positions);
     let rhs_colored = apply_colors(&rhs_src, false, &rhs_positions);
 
-    let lang_name = match &lang {
-        Some(lang) => lang.name.clone(),
-        None => "plain text".to_string(),
-    };
-    println!("{}", style::header(rhs_path, &lang_name));
     print!(
         "{}",
         apply_groups(
             &lhs_colored,
             &rhs_colored,
             &groups,
+            &lhs_matched_lines,
             lhs_content_width,
             lhs_column_width,
             rhs_column_width,
